from abc import ABCMeta, abstractmethod


class _DatabaseCruncher(metaclass=ABCMeta):
    """
    Base class for database crunching.

    Common operations are shared here, allowing subclasses to just focus on
    implementing the crunching algorithms.
    """

    def __init__(self, db):
        """
        Initialise the database cruncher

        Parameters
        ----------
        db : IamDataFrame
            The database to use
        """
        self._db = db.copy()

    @abstractmethod
    def derive_relationship(self, variable_follower, variable_leaders, **kwargs):
        """
        Derive the relationship between two variables from the database

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|C5F12"``).

        variable_leaders : list[str]
            The variables we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2", "Emissions|CH4"]``)

        **kwargs
            Keyword arguments used by this class to derive the relationship between
            ``variable_follower`` and ``variable_leaders``.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.
        """
        # TODO: think about how to add region handling in here...

    def _check_follower_and_leader_in_db(self, variable_follower, variable_leaders):

        if not all([v in self._db.variables().tolist() for v in variable_leaders]):
            error_msg = "No data for `variable_leaders` ({}) in database".format(
                variable_leaders
            )
            raise ValueError(error_msg)

        if variable_follower not in self._db.variables().tolist():
            error_msg = "No data for `variable_follower` ({}) in database".format(
                variable_follower
            )
            raise ValueError(error_msg)



"""
Module for the database cruncher which makes a linear interpolator between known values
"""

from pyam import IamDataFrame

from ..utils import _get_unit_of_variable, _make_interpolator, _make_wide_db
from .base import _DatabaseCruncher


class DatabaseCruncherLinearInterpolation(_DatabaseCruncher):
    """
    Database cruncher which uses linear interpolation.

    This cruncher derives the relationship between two variables by simply linearly
    interpolating between values in the cruncher database. It does not do any smoothing
    and is best-suited for smaller databases.

    In the case where there is more than one value of the follower variable for a
    given value of the leader variable, the average will be used. For example, if
    one scenario has CH4 emissions of 10 MtCH4/yr whilst another has CH4
    emissions of 20 MtCH4/yr in 2020 whilst both scenarios have CO2 emissions
    of exactly 15 GtC/yr in 2020, the interpolation will use the average value from the
    two scenarios i.e. 15 Mt CH4/yr.

    Beyond the bounds of input data, the linear interpolation is held constant.
    For example, if the maximum CO2 emissions in 2020 in the database is
    25 GtC/yr, and CH4 emissions for this level of CO2 emissions are 15 MtCH4/yr,
    then even if we infill using a CO2 emissions value of 100 GtC/yr in 2020, the
    returned CH4 emissions will be 15 MtCH4/yr.
    """

    def derive_relationship(self, variable_follower, variable_leaders):
        """
        Derive the relationship between two variables from the database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CH4"``).

        variable_leaders : list[str]
            The variable(s) we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``).

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.

        Raises
        ------
        ValueError
            There is no data of the appropriate type in the database.
        """
        if len(variable_leaders) != 1:
            raise NotImplementedError(
                "Having more than one `variable_leaders` is not yet implemented"
            )
        use_db = self._db.filter(variable=[variable_leaders[0], variable_follower])
        if use_db.data.empty:
            raise ValueError(
                "There is no data of the appropriate type in the database."
            )
        leader_units = _get_unit_of_variable(use_db, variable_leaders)
        follower_units = _get_unit_of_variable(use_db, variable_follower)
        if len(leader_units) == 0:
            raise ValueError(
                "No data for `variable_leaders` ({}) in database".format(
                    variable_leaders
                )
            )
        if len(follower_units) == 0:
            raise ValueError(
                "No data for `variable_follower` ({}) in database".format(
                    variable_follower
                )
            )
        leader_units = leader_units[0]
        use_db_time_col = use_db.time_col
        use_db = _make_wide_db(use_db)
        interpolators = _make_interpolator(
            variable_follower, variable_leaders, use_db, use_db_time_col
        )

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`DatabaseCruncherSSPSpecificRelation`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                The key db_times for filling are not in ``in_iamdf``.
            """
            if use_db_time_col != in_iamdf.time_col:
                raise ValueError(
                    "`in_iamdf` time column must be the same as the time column used "
                    "to generate this filler function (`{}`)".format(use_db_time_col)
                )

            var_units = _get_unit_of_variable(in_iamdf, variable_leaders)
            if var_units.size == 0:
                raise ValueError(
                    "There is no data for {} so it cannot be infilled".format(
                        variable_leaders
                    )
                )
            var_units = var_units[0]
            lead_var = in_iamdf.filter(variable=variable_leaders)
            assert (
                lead_var["unit"].nunique() == 1
            ), "There are multiple units for the lead variable."
            if var_units != leader_units:
                raise ValueError(
                    "Units of lead variable is meant to be `{}`, found `{}`".format(
                        leader_units, var_units
                    )
                )
            times_needed = set(in_iamdf.data[in_iamdf.time_col])
            if any(x not in interpolators.keys() for x in times_needed):
                raise ValueError(
                    "Not all required timepoints are present in the database we "
                    "crunched, we crunched \n\t`{}`\nbut you passed in \n\t{}".format(
                        list(interpolators.keys()),
                        in_iamdf.timeseries().columns.tolist(),
                    )
                )
            output_ts = lead_var.timeseries()
            for time in times_needed:
                output_ts[time] = interpolators[time](output_ts[time])
            output_ts.reset_index(inplace=True)
            output_ts["variable"] = variable_follower
            output_ts["unit"] = follower_units[0]
            return IamDataFrame(output_ts)

        return filler



"""
Module for the database cruncher which uses the 'closest RMS' technique.
"""
import warnings

import pyam

from ..utils import _get_unit_of_variable
from .base import _DatabaseCruncher


class DatabaseCruncherRMSClosest(_DatabaseCruncher):
    """
    Database cruncher which uses the 'closest RMS' technkque.

    This cruncher derives the relationship between two variables by finding the
    scenario which has the closest lead gas timeseries in the database.  The follower
    gas timeseries is then simply copied from the closest scenario.

    Here, 'closest' is defined as the smallest time-averaged root mean squared (L2)
    difference.

    .. math::
        RMS = \\left ( \\frac{1}{n} \\sum_{t=0}^n (E_l(t) - e_l(t))^2 \\right )^{1/2}

    where :math:`n` is the total number of timesteps in the lead gas' timeseries,
    :math:`E_l(t)` is the lead gas emissions timeseries and :math:`e_l(t)` is a lead
    gas emissions timeseries in the infiller database.
    """

    def derive_relationship(self, variable_follower, variable_leaders):
        """
        Derive the relationship between two variables from the database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|C5F12"``).

        variable_leaders : list[str]
            The variable we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``).

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.

        Raises
        ------
        ValueError
            ``variable_leaders`` contains more than one variable.

        ValueError
            There is no data for ``variable_leaders`` or ``variable_follower`` in the
            database.
        """
        self._check_iamdf_lead(variable_leaders)
        iamdf_follower = self._get_iamdf_section(variable_follower)
        data_follower_time_col = iamdf_follower.time_col
        iamdf_lead = self._db.filter(variable=variable_leaders)
        iamdf_lead, iamdf_follower = _filter_for_overlap(
            iamdf_lead, iamdf_follower, ["scenario", "model", data_follower_time_col]
        )

        leader_unit = _get_unit_of_variable(iamdf_lead, variable_leaders)
        leader_unit = leader_unit[0]

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`DatabaseCruncherRMSClosest`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                If there are any inconsistencies between the timeseries, units or
                expectations of the program and ``in_iamdf``, compared to the database
                used to generate this ``filler`` function.
            """
            lead_var = in_iamdf.filter(variable=variable_leaders)

            var_units = lead_var["unit"].unique()
            if len(var_units) != 1:
                raise ValueError("More than one unit detected for input timeseries")

            var_units = var_units[0]
            if var_units != leader_unit:
                raise ValueError(
                    "Units of lead variable is meant to be {}, found {}".format(
                        leader_unit, var_units
                    )
                )

            if data_follower_time_col != in_iamdf.time_col:
                raise ValueError(
                    "`in_iamdf` time column must be the same as the time column used "
                    "to generate this filler function (`{}`)".format(
                        data_follower_time_col
                    )
                )

            key_timepoints_filter_iamdf = {
                data_follower_time_col: list(set(lead_var[data_follower_time_col]))
            }
            key_timepoints_filter_lead = {
                data_follower_time_col: list(set(iamdf_lead[data_follower_time_col]))
            }

            def get_values_at_key_timepoints(idf, time_filter):
                # filter warning about empty data frame as we handle it ourselves
                to_return = idf.filter(**time_filter)
                if to_return.data.empty:
                    raise ValueError(
                        "No time series overlap between the original and unfilled data"
                    )
                return to_return

            lead_var_timeseries = get_values_at_key_timepoints(
                lead_var, key_timepoints_filter_lead
            ).timeseries()
            iamdf_lead_timeseries = get_values_at_key_timepoints(
                iamdf_lead, key_timepoints_filter_iamdf
            ).timeseries()

            output_ts_list = []
            for label, row in lead_var_timeseries.iterrows():
                closest_ts = _select_closest(iamdf_lead_timeseries, row)

                # Filter to find the matching follow data for the same model, scenario
                # and region
                tmp = iamdf_follower.filter(
                    model=closest_ts["model"], scenario=closest_ts["scenario"]
                ).data

                # Update the model and scenario to match the elements of the input.
                tmp["model"] = label[lead_var_timeseries.index.names.index("model")]
                tmp["scenario"] = label[
                    lead_var_timeseries.index.names.index("scenario")
                ]
                output_ts_list.append(tmp)

            return pyam.concat(output_ts_list)

        return filler

    def _check_iamdf_lead(self, variable_leaders):
        if len(variable_leaders) > 1:
            raise ValueError(
                "For `DatabaseCruncherRMSClosest`, ``variable_leaders`` should only "
                "contain one variable"
            )

        if not all([v in self._db.variables().tolist() for v in variable_leaders]):
            error_msg = "No data for `variable_leaders` ({}) in database".format(
                variable_leaders
            )
            raise ValueError(error_msg)

    def _get_iamdf_section(self, variables):
        # filter warning about empty data frame as we handle it ourselves
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            iamdf_section = self._db.filter(variable=variables)

        data_section = iamdf_section.data
        if data_section.empty:
            error_msg = "No data for `variable_follower` ({}) in database".format(
                variables
            )
            raise ValueError(error_msg)

        return iamdf_section


def _select_closest(to_search_df, target_series):
    """
    Find row in ``to_search_df`` that is closest to the target array.

    Here, 'closest' is in the root-mean squared sense. In the event that multiple rows
    are equally close, returns first row.

    Parameters
    ----------
    to_search_df : :obj:`pd.DataFrame`
        The rows of this dataframe are the candidate closest vectors

    target_series : :obj:`pd.Series`
        The vector to which we want to be close

    Returns
    -------
    dict
        Metadata of the closest row.
    """
    if len(target_series.shape) != 1:
        raise ValueError("Target array is multidimensional")

    if target_series.shape[0] != to_search_df.shape[1]:
        raise ValueError(
            "Target array does not match the size of the searchable arrays"
        )

    closeness = []
    for label, row in to_search_df.iterrows():
        rms = (((target_series - row) ** 2).mean()) ** 0.5
        closeness.append((label, rms))

    # Find the minimum closeness and return the index of it
    labels, rmss = list(zip(*closeness))
    to_return = rmss.index(min(rmss))
    return dict(zip(to_search_df.index.names, labels[to_return]))


def _filter_for_overlap(df1, df2, cols):
    """
    Returns rows in the two input dataframes which have the same columns
    Parameters
    ----------
    df1 : :obj:`pd.DataFrame`
        The first dataframe (order is irrelevant)
    df2 : :obj:`pd.DataFrame`
        The second dataframe (order is irrelevant)
    cols: list[str]
        List of columns that should be identical between the two dataframes.
    Returns
    -------
    (:obj:`pd.DataFrame`, :obj:`pd.DataFrame`)
        The two dataframes in the order they were put in, now filtered for some columns
        being identical.
    """
    lead_data = df1.data.set_index(cols)
    follow_data = df2.data.set_index(cols)
    shared_indices = [ind for ind in lead_data.index if ind in follow_data.index]
    if shared_indices:
        lead_data = lead_data.loc[shared_indices]
        follow_data = follow_data.loc[shared_indices]
        return pyam.IamDataFrame(lead_data), pyam.IamDataFrame(follow_data)
    raise ValueError("No model/scenario overlap between leader and follower data")



"""Utility plotting functions"""
import itertools
import os.path

import matplotlib.pyplot as plt
import numpy as np

from .stats import rolling_window_find_quantiles


def _plot_emission_correlations_quantile_rolling_windows(
    emms_df,
    output_dir,
    years,
    x_gas="Emissions|CO2",
    quantiles=None,
    quantile_boxes=20,
    quantile_decay_factor=1,
    model_colours=True,
    legend_fraction=0.65,
    region="World",
):
    """
    Calculates the relationship between different sorts of emissions at pre-specified
    times. Plots graphs of the relationships between each of the variables and
    optionally plots given quantiles on top of this.
    Saves csv files of the correlation coefficients and the rank correlation
    coefficients between emissions at specified locations.

    Parameters
    ----------
    emms_df : :obj:`pyam.IamDataFrame`
        The database to search for correlations between named values

    output_dir : str
        The folder location to save the files.

    years : list[int]
        The years upon which to calculate correlations.

    x_gas : str
        The name of the gas to plot on the x-axis.

    quantiles : list[float], None
        If not none, the function will also calculate the quantiles specified by this
        list, using rolling windows as documented in rolling_window_find_quantiles. If
        none, the following four values are irrelevant.

    quantile_boxes : int, None
        The number of points at which quantiles should be evaluated. For details see
        rolling_window_find_quantiles documentation.

    quantile_decay_factor : float
        This determines how strong the local weighting is for the quantiles. 1 is the
        standard value. For details see rolling_window_find_quantiles documentation.

    model_colours : bool
        If true, the plot of quantiles will include a legend and differently coloured
        trends.

    legend_fraction : float
        The size of the legend, if plotted.
    """
    for year_of_interest in years:
        # Obtain the list of gases to examine
        df_gases = (
            emms_df.filter(region=region, year=year_of_interest, level=1)
            .filter(variable="Emissions|*")
            .variables(True)
            .set_index("variable")
        )

        if x_gas not in emms_df.variables().values:
            raise ValueError("No {} data".format(x_gas))

        # Check that the list has only one entry for each gas
        assert not any(df_gases.index.duplicated()), "Index contains duplicated entries"
        x_units = emms_df.filter(variable=x_gas)["unit"].iloc[0]

        for y_gas_ind in range(df_gases.count()[0]):
            plt.close()
            y_gas = df_gases.index[y_gas_ind]
            y_units = df_gases.get("unit")[y_gas_ind]

            # Create the dataframe to plot
            seaborn_df = emms_df.filter(
                variable=[y_gas, x_gas], region=region, year=year_of_interest
            ).pivot_table(
                ["year", "model", "scenario", "region"], ["variable"], aggfunc="mean"
            )

            # Cleaning the data
            seaborn_df.replace(r"^\s*$", np.nan, regex=True, inplace=True)
            seaborn_df = seaborn_df.dropna().reset_index()
            seaborn_df.loc[:, [y_gas, x_gas]] = seaborn_df[[y_gas, x_gas]].astype(float)

            # Plot the results
            if model_colours:
                _plot_multiple_models(
                    legend_fraction, seaborn_df, x_gas, y_gas, x_units, y_units
                )
            # if all plots are the same colour, we don't have to do all this work
            else:
                _plot_emissions(
                    legend_fraction, seaborn_df, x_gas, y_gas, x_units, y_units
                )

            # Optionally calculate and plot quantiles
            if quantiles is not None:
                quant_df = rolling_window_find_quantiles(
                    seaborn_df[x_gas],
                    seaborn_df[y_gas],
                    quantiles,
                    quantile_boxes,
                    quantile_decay_factor,
                )
                plt.plot(quant_df.index, quant_df)

                if not model_colours:
                    plt.legend(
                        quant_df.keys(), loc="center left", bbox_to_anchor=(1, 0.5)
                    )

                if output_dir is not None:
                    quant_df.to_csv(
                        os.path.join(
                            output_dir,
                            "{}_{}_{}.csv".format(
                                x_gas.split("|")[-1],
                                y_gas.split("|")[-1],
                                year_of_interest,
                            ),
                        )
                    )

            # Report the results
            if output_dir is not None:
                plt.savefig(
                    os.path.join(
                        output_dir,
                        "{}_{}_{}.png".format(
                            x_gas.split("|")[-1], y_gas.split("|")[-1], year_of_interest
                        ),
                    )
                )
            print("Finished {} vs {} in {}".format(x_gas, y_gas, year_of_interest))


def _plot_emissions(legend_fraction, seaborn_df, x_gas, y_gas, x_units, y_units):
    ax = plt.subplot(111)
    colours_for_plot = "black"
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * legend_fraction, box.height])
    plt.scatter(x=x_gas, y=y_gas, color=colours_for_plot, data=seaborn_df, alpha=0.5)
    plt.xlabel("Emissions of {} ({})".format(x_gas.split("|")[-1], x_units))
    plt.ylabel("Emissions of {} ({})".format(y_gas.split("|")[-1], y_units))


def _plot_multiple_models(legend_fraction, seaborn_df, x_gas, y_gas, x_units, y_units):
    ax = plt.subplot(111)
    all_models = list(seaborn_df["model"].unique())
    markers = itertools.cycle(["s", "o", "v", "<", ">", ","])
    for model in all_models:
        to_plot = np.where(seaborn_df["model"] == model)[0]
        if any(to_plot):
            plt.scatter(
                x=seaborn_df[x_gas].loc[to_plot],
                y=seaborn_df[y_gas].loc[to_plot],
                label=model,
                alpha=0.5,
                marker=next(markers),
            )
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * legend_fraction, box.height])
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    plt.xlabel("Emissions of {} ({})".format(x_gas.split("|")[-1], x_units))
    plt.ylabel("Emissions of {} ({})".format(y_gas.split("|")[-1], y_units))


def _plot_reconstruct_value_with_cruncher(
    var_inst, var_units, interp_values, originals, crunchers_name, save_plots, db_all
):
    """
    Plots the relationships between two variables in the database at a particular time,
    and how a cruncher reconstructs a particular line.
    Parameters
    ----------
    var_inst : str
        Name of the variable being plotted

    var_units : str
        Name of the units of the variable

    interp_values : Series
        The data for the crunched values, x values as the index

    originals : Series
        The full dataset of all input values, x values as the index

    crunchers_name : str
        Name of the cruncher.

    save_plots : str
        The dirctory in which the plots are to be saved.
    """
    plt.close()
    plt.subplot(111)
    plt.scatter(x=originals.index, y=originals, label="True values", alpha=0.8)
    plt.scatter(
        x=interp_values.index,
        y=interp_values,
        label="Model values",
        marker="s",
        alpha=0.5,
    )
    plt.xlabel("Year")
    plt.ylabel("Emissions of {} ({})".format(var_inst, var_units[0]))
    to_plot = db_all.filter(variable=var_inst)
    plt.scatter(
        to_plot["year"],
        to_plot["value"],
        label="Other values",
        alpha=0.5,
        marker="v",
        s=8,
    )
    plt.legend()
    plt.savefig(
        save_plots + "{}_{}.png".format(crunchers_name, var_inst.split("|")[-1])
    )



from .base import _DatabaseCruncher
from .linear_interpolation import DatabaseCruncherLinearInterpolation


class DatabaseCruncherScenarioAndModelSpecificInterpolate(_DatabaseCruncher):
    """
    Database cruncher which pre-filters to only use data from specific scenarios, then
    runs the linear interpolator to return values from that set of scenarios. See the
    documentation of DatabaseCruncherLinearInterpolation for more details.
    """

    def derive_relationship(
        self,
        variable_follower,
        variable_leaders,
        required_scenario="*",
        required_model="*",
    ):
        """
        Derive the relationship between two variables from the database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CH4"``).

        variable_leaders : list[str]
            The variable(s) we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``).

        required_scenario : str or list[str]
            The string(s) which all relevant scenarios are required to match. This may
            have *s to represent wild cards. It defaults to "*" to accept all scenarios.

        required_model : str or list[str]
            The string(s) which all relevant models are required to match. This may have
            *s to represent wild cards. It defaults to "*" to accept all models.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.

        Raises
        ------
        ValueError
            There is no data of the appropriate type in the database.
             There may be a typo in the SSP option.
        """
        use_db = self._db.filter(scenario=required_scenario, model=required_model)
        if use_db.data.empty:
            raise ValueError(
                "There is no data of the appropriate type in the database."
                " There may be a typo in the SSP option."
            )
        cruncher = DatabaseCruncherLinearInterpolation(use_db)
        return cruncher.derive_relationship(variable_follower, variable_leaders)



import logging

from silicone.utils import _construct_consistent_values

logger = logging.getLogger(__name__)


def infill_composite_values(df, composite_dic=None):
    """
    Constructs a series of aggregate variables, calculated as the sums of variables
    that have been reported. If given factors terms too, the terms will be multiplied
    by the factors before summing.

    Parameters
    ----------
    df : :obj:`pyam.IamDataFrame`
        Input data from which to construct consistent values. This is assumed to be
        fully infilled. This will not be checked.

    composite_dic : dict {str: list[str] or dict{str: float}}
        Key: The variable names of the composite. Value: The variable names of the
        constituents, which may include wildcards ('*'). Optionally, these may be
        dictionaries of the names to factors, which we multiply the numbers by before
        summing them. Defaults to a list of PFC, HFC, F-Gases, CO2 and Kyoto gases.

    Returns
    -------
    :obj:`pyam.IamDataFrame`
        :obj:`pyam.IamDataFrame` containing the composite values.
    """
    if composite_dic is None:
        composite_dic = {
            "Emissions|PFC": [
                "Emissions*|CF4",
                "Emissions*|C2F6",
                "Emissions*|C3F8",
                "Emissions*|C4F10",
                "Emissions*|C5F12",
                "Emissions*|C6F14",
                "Emissions*|C7F16",
                "Emissions*|C8F18",
            ],
            "Emissions|HFC": ["Emissions|HFC*"],
            "Emissions|F-Gases": ["Emissions|PFC", "Emissions|HFC", "Emissions|SF6",],
            "Emissions|Kyoto Gases (AR5-GWP100)": [
                "Emissions|CO2",
                "Emissions|CH4",
                "Emissions|N2O",
                "Emissions|F-Gases",
            ],
        }

    # Do a simple test to ensure we have any relevant data. This does not prevent
    # problems from missing subsets of data.
    to_delete = []

    # Disable core logging to prevent warnings about empty filters
    logging.getLogger("pyam.core").setLevel(logging.CRITICAL)
    for composite, composite_variables in composite_dic.items():
        if df.filter(variable=composite_variables).data.empty:
            logger.warning("No data found for {}".format(composite_variables))
            to_delete.append(composite)

    logging.getLogger("pyam.core").setLevel(logging.WARNING)
    for x in to_delete:
        composite_dic.pop(x)

    # If the composite variables already exist, remove them first
    df = df.filter(variable=composite_dic.keys(), keep=False)
    for composite, composite_variables in composite_dic.items():
        if isinstance(composite_variables, dict):
            temp_df = df.filter(variable=[composite] + list(composite_variables.keys()))
            for contributor, factor in composite_variables.items():
                temp_df.data.loc[
                    temp_df.data["variable"] == contributor, "value"
                ] *= factor

            composite_df = _construct_consistent_values(
                composite, composite_variables, temp_df
            )

        else:
            composite_df = _construct_consistent_values(
                composite, composite_variables, df
            )
        try:
            to_return.append(composite_df)
        except NameError:
            to_return = composite_df

    return to_return



"""
Uses the 'time-dependent ratio' database cruncher designed for constructing an
aggregate variable and breaking this mix into its constituents.
"""

import pyam

from silicone.database_crunchers import DatabaseCruncherTimeDepRatio
from silicone.utils import convert_units_to_MtCO2_equiv


class DecomposeCollectionTimeDepRatio:
    """
    Constructs an aggregate variable and uses the 'time-dependent ratio' technique to
    calculate what this predicts for our database.
    """

    def __init__(self, db):
        """
        Initialises the database to use for infilling.

        Parameters
        ----------
        db : IamDataFrame
            The database for infilling.
        """
        self._db = db.copy()

    def _construct_consistent_values(self, aggregate_name, components, db_to_generate):
        """
            Calculates the sum of the components and creates an IamDataFrame with this
            value under variable type `aggregate_name`.

            Parameters
            ----------
            aggregate_name : str
                The name of the aggregate variable.

            components : [str]
                List of the names of the variables to be summed.

            db_to_generate : :obj:`pyam.IamDataFrame`
                Input data from which to construct consistent values.

            Return
            ------
            :obj:`pyam.IamDataFrame`
                Consistently calculated aggregate data.
            """
        assert (
            aggregate_name not in db_to_generate.variables().values
        ), "We already have a variable of this name"
        relevant_db = db_to_generate.filter(variable=components)
        units = relevant_db.data["unit"].drop_duplicates().sort_values()
        unit_equivs = units.map(lambda x: x.replace("-equiv", "")).drop_duplicates()
        if len(unit_equivs) == 0:
            raise ValueError(
                "Attempting to construct a consistent {} but none of the components "
                "present".format(aggregate_name)
            )
        elif len(unit_equivs) > 1:
            raise ValueError(
                "Too many units found to make a consistent {}".format(aggregate_name)
            )
        use = (
            relevant_db.data.groupby(
                ["model", "scenario", "region", relevant_db.time_col]
            )
            .agg("sum")
            .reset_index()
        )
        # Units are sorted in alphabetical order so we choose the first to get -equiv
        use["unit"] = units.iloc[0]
        use["variable"] = aggregate_name
        return pyam.IamDataFrame(use)

    def _set_of_units_without_equiv(self, df):
        """
        Parameters
        ----------
        df : obj:`pyam.IamDataFrame`
            The dataframe whose units we want

        Returns
        -------
        Set(str)
            The set of units from the dataframe with "-equiv" removed
        """
        return set(df.data["unit"].map(lambda x: x.replace("-equiv", "")))

    def infill_components(
        self, aggregate, components, to_infill_df, use_ar4_data=False
    ):
        """
        Derive the relationship between the composite variables and their sum, then use
        this to deconstruct the sum.

        Parameters
        ----------
        aggregate : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CO2"``). Unlike in most crunchers, we do not expect the
            database to already contain this data.

        components : list[str]
            The variables whose sum should be equal to the timeseries of the aggregate
             (e.g. ``["Emissions|CO2|AFOLU", "Emissions|CO2|Energy"]``).

        to_infill_df : :obj:`pyam.IamDataFrame`
            The dataframe that already contains the ``aggregate`` variable, but needs
            the ``components`` to be infilled.

        use_ar4_data : bool
            If true, we convert all values to Mt CO2 equivalent using the IPCC AR4
            GWP100 data, otherwise (by default) we use the GWP100 data from AR5.

        Returns
        -------
        :obj:`pyam.IamDataFrame`
            The infilled data resulting from the calculation.

        Raises
        ------
        ValueError
            There is no data for ``variable_leaders`` or ``variable_follower`` in the
            database.
        """
        assert (
            aggregate in to_infill_df.variables().values
        ), "The database to infill does not have the aggregate variable"
        assert all(
            y not in components for y in to_infill_df.variables().values
        ), "The database to infill already has some component variables"
        assert len(to_infill_df.data.columns) == len(self._db.data.columns) and all(
            to_infill_df.data.columns == self._db.data.columns
        ), (
            "The database and to_infill_db fed into this have inconsistent columns, "
            "which will prevent adding the data together properly."
        )
        self._db.filter(variable=components, inplace=True)
        # We only want to reference cases where all the required components are found
        combinations = self._db.data[["model", "scenario", "region"]].drop_duplicates()
        for ind in range(len(combinations)):
            model, scenario, region = combinations.iloc[ind]
            found_vars = self._db.filter(
                model=model, scenario=scenario, region=region
            ).variables()
            if any(comp not in found_vars.values for comp in components):
                self._db.filter(
                    model=model, scenario=scenario, keep=False, inplace=True
                )
        if len(self._set_of_units_without_equiv(self._db)) > 1:
            db_to_generate = convert_units_to_MtCO2_equiv(
                self._db, use_ar4_data=use_ar4_data
            )
        else:
            db_to_generate = self._db
        consistent_composite = self._construct_consistent_values(
            aggregate, components, db_to_generate
        )
        self._db.append(consistent_composite, inplace=True)
        cruncher = DatabaseCruncherTimeDepRatio(self._db)
        if self._set_of_units_without_equiv(
            to_infill_df
        ) != self._set_of_units_without_equiv(consistent_composite):
            raise ValueError(
                "The units of the aggregate variable are inconsistent between the "
                "input and constructed data. We input {} and constructed {}.".format(
                    self._set_of_units_without_equiv(to_infill_df),
                    self._set_of_units_without_equiv(consistent_composite),
                )
            )
        for leader in components:
            to_add = cruncher.derive_relationship(leader, [aggregate])(to_infill_df)
            try:
                df_to_append.append(to_add, inplace=True)
            except NameError:
                df_to_append = to_add
        return df_to_append



"""
Module for the database cruncher which uses the 'time-dependent ratio' technique.
"""

import warnings

import numpy as np
import pandas as pd
from pyam import IamDataFrame

from .base import _DatabaseCruncher


class DatabaseCruncherTimeDepRatio(_DatabaseCruncher):
    """
    Database cruncher which uses the 'time-dependent ratio' technique.

    This cruncher derives the relationship between two variables by simply assuming
    that the follower timeseries is equal to the lead timeseries multiplied by a
    time-dependent scaling factor. The scaling factor is the ratio of the
    follower variable to the lead variable. If the database contains many such pairs,
    the scaling factor is the ratio between the means of the values. By default, the
    calculation will include only values where the lead variable takes the same sign
    (+ or -) in the infilling database as in the case infilled. This prevents getting
    negative values of emissions that cannot be negative. To allow cases where we
    have no data of the correct sign, set `same_sign = False` in `derive_relationship`.

    Once the relationship is derived, the 'filler' function will infill following:

    .. math::
        E_f(t) = R(t) * E_l(t)

    where :math:`E_f(t)` is emissions of the follower variable and :math:`E_l(t)` is
    emissions of the lead variable.

    :math:`R(t)` is the scaling factor, calculated as the ratio of the means of the
    the follower and the leader in the infiller database, denoted with
    lower case e. By default, we include only cases where `sign(e_l(t))` is the same in
    both databases).

    .. math::
        R(t) = \\frac{mean( e_f(t) )}{mean( e_l(t) )})

    """

    def derive_relationship(self, variable_follower, variable_leaders, same_sign=True):
        """
        Derive the relationship between two variables from the database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|C5F12"``).

        variable_leaders : list[str]
            The variable we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``).

        same_sign : bool
            Do we want to only use data where the leader has the same sign in the
            infiller and infillee data? If so, we have a potential error from
            not having data of the correct sign, but have more confidence in the
            sign of the follower data.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.

        Raises
        ------
        ValueError
            ``variable_leaders`` contains more than one variable.

        ValueError
            There is no data for ``variable_leaders`` or ``variable_follower`` in the
            database.
        """
        iamdf_follower, data_follower = self._get_iamdf_followers(
            variable_follower, variable_leaders
        )

        data_follower_unit = np.unique(iamdf_follower.data["unit"].values)
        if data_follower_unit.size == 1:
            data_follower_unit = data_follower_unit[0]
        else:
            raise ValueError("There are multiple/no units in follower data")
        data_follower_time_col = iamdf_follower.time_col
        iamdf_leader = self._db.filter(variable=variable_leaders[0])
        data_leader = iamdf_leader.timeseries()
        if iamdf_leader["unit"].nunique() != 1:
            raise ValueError("There are multiple/no units for the leader data.")
        if data_follower.size != data_leader.size:
            error_msg = "The follower and leader data have different sizes"
            raise ValueError(error_msg)
        # Calculate the ratios to use
        all_times = np.unique(iamdf_leader.data[iamdf_leader.time_col])
        scaling = pd.DataFrame(index=all_times, columns=["pos", "neg"])
        if same_sign:
            # We want to have separate positive and negative answers. We calculate a
            # tuple, first for positive and then negative values.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                for year in all_times:
                    pos_inds = data_leader[year].values > 0
                    scaling["pos"][year] = np.nanmean(
                        data_follower[year].iloc[pos_inds].values
                    ) / np.nanmean(data_leader[year].iloc[pos_inds].values)
                    scaling["neg"][year] = np.nanmean(
                        data_follower[year].iloc[~pos_inds].values
                    ) / np.nanmean(data_leader[year].iloc[~pos_inds].values)
        else:
            # The tuple is the same in both cases
            for year in all_times:
                scaling["pos"][year] = np.mean(data_follower[year].values) / np.mean(
                    data_leader[year].values
                )
            scaling["neg"] = scaling["pos"]

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`DatabaseCruncherTimeDepRatio`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled-in data (without original source data)

            Raises
            ------
            ValueError
                The key year for filling is not in ``in_iamdf``.
            """
            lead_var = in_iamdf.filter(variable=variable_leaders)
            assert (
                lead_var["unit"].nunique() == 1
            ), "There are multiple units for the lead variable."
            if data_follower_time_col != in_iamdf.time_col:
                raise ValueError(
                    "`in_iamdf` time column must be the same as the time column used "
                    "to generate this filler function (`{}`)".format(
                        data_follower_time_col
                    )
                )
            times_needed = set(in_iamdf.data[in_iamdf.time_col])
            if any(
                [
                    k not in set(iamdf_follower[data_follower_time_col])
                    for k in times_needed
                ]
            ):
                error_msg = (
                    "Not all required timepoints are in the data for "
                    "the lead gas ({})".format(variable_leaders[0])
                )
                raise ValueError(error_msg)
            output_ts = lead_var.timeseries()

            for year in times_needed:
                if (
                    scaling.loc[year][
                        output_ts[year].map(lambda x: "neg" if x < 0 else "pos")
                    ]
                    .isnull()
                    .values.any()
                ):
                    raise ValueError(
                        "Attempt to infill {} data using the time_dep_ratio cruncher "
                        "where the infillee data has a sign not seen in the infiller "
                        "database for year "
                        "{}.".format(variable_leaders, year)
                    )
                output_ts[year] = (
                    output_ts[year].values
                    * scaling.loc[year][
                        output_ts[year].map(lambda x: "pos" if x > 0 else "neg")
                    ].values
                )
            output_ts.reset_index(inplace=True)
            output_ts["variable"] = variable_follower
            output_ts["unit"] = data_follower_unit

            return IamDataFrame(output_ts)

        return filler

    def _get_iamdf_followers(self, variable_follower, variable_leaders):
        if len(variable_leaders) > 1:
            raise ValueError(
                "For `DatabaseCruncherTimeDepRatio`, ``variable_leaders`` should only "
                "contain one variable"
            )

        self._check_follower_and_leader_in_db(variable_follower, variable_leaders)

        iamdf_follower = self._db.filter(variable=variable_follower)
        data_follower = iamdf_follower.timeseries()

        return iamdf_follower, data_follower



"""
Module for the database cruncher which uses the 'constant given ratio' technique.
"""
import logging

import pandas as pd
from pyam import IamDataFrame

from .base import _DatabaseCruncher

logger = logging.getLogger(__name__)


class DatabaseCruncherConstantRatio(_DatabaseCruncher):
    """
    Database cruncher which uses the 'constant given ratio' technique.

    This cruncher does not require a database upon initialisation. Instead, it requires
    a constant and a unit to be input when deriving relations. This constant,
    :math:`s`, is the ratio of the follower variable to the lead variable i.e.:

    .. math::

        E_f(t) = s * E_l(t)

    where :math:`E_f(t)` is emissions of the follower variable and :math:`E_l(t)` is
    emissions of the lead variable.
    """

    def __init__(self, db=None):
        """
        Initialise the database cruncher

        Parameters
        ----------
        db : IamDataFrame
            Supplied to ensure consistency with the base class. This cruncher doesn't
            actually use the database at all.
        """
        if db is not None:
            logger.info(
                "%s won't use any information from the database", self.__class__
            )

        super().__init__(pd.DataFrame())

    def derive_relationship(self, variable_follower, variable_leaders, ratio, units):
        """
        Derive the relationship between two variables from the database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|C5F12"``).

        variable_leaders : list[str]
            The variable we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``).

        ratio : float
            The ratio between the leader and the follower data

        units : str
            The units of the follower data.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
        """
        if len(variable_leaders) > 1:
            raise ValueError(
                "For `DatabaseCruncherConstantRatio`, ``variable_leaders`` should only "
                "contain one variable"
            )

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`DatabaseCruncherTimeDepRatio`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled-in data (without original source data)

            Raises
            ------
            ValueError
                The key year for filling is not in ``in_iamdf``.
            """
            output_ts = in_iamdf.filter(variable=variable_leaders)
            assert (
                output_ts["unit"].nunique() == 1
            ), "There are multiple units for the lead variable."
            output_ts["value"] = output_ts["value"] * ratio
            output_ts["variable"] = variable_follower
            output_ts["unit"] = units

            return output_ts

        return filler



import datetime as dt
import logging
import os.path

import numpy as np
import pandas as pd
import pyam
import scipy.interpolate
from openscm_units.unit_registry import ScmUnitRegistry
from pint.errors import DimensionalityError

logger = logging.getLogger(__name__)

# initialise our own registry to avoid conflicts
_ur = ScmUnitRegistry()
_ur.add_standards()

"""
Utils contains a number of helpful functions that don't belong elsewhere.
"""


def find_matching_scenarios(
    options_df,
    to_compare_df,
    variable_follower,
    variable_leaders,
    classify_scenarios,
    classify_models=["*"],
    return_all_info=False,
    use_change_not_abs=False,
):
    """
    Groups scenarios and models into different classifications and uses those to
    work out which group contains a trendline most similar to the data. These
    combinations may group several models/scenarios together by means of wild cards.
    Most similar means having the smallest total squared distance between the
    to_compare_df value of variable_follower and the variable_follower values
    interpolated in options_df at the variable_leaders points in to_compare_df, i.e.
    assuming errors only exist in variable_follower.
    In the event of a tie between different scenario/model classifications, it returns the
    scenario/model combination that occurs earlier in the input lists, looping through
    scenarios first.

    Parameters
    ----------
    options_df : :obj:`pyam.IamDataFrame`
        The dataframe containing the data for each scenario and model

    to_compare_df : :obj:`pyam.IamDataFrame`
        The dataframe we wish to find the scenario group closest to. May contain one
        or more scenarios, we minimise the least squared errors for all the data
        colleectively.

    variable_follower : str
        The variable we want to interpolate and compare to the value in to_compare_df

    variable_leaders : list[str]
        The variable(s) we want to use to construct the interpolation
        (e.g. ``["Emissions|CO2"]``). In the event that there are multiple, we
        interpolate with each one separately and minimise the sum of the squared
        errors.

    classify_scenarios : list[str]
        The names of scenarios or groups of scenarios that are possible matches.
        This may have "\*"s to represent wild cards, hence multiple scenarios will have
        all their data combined to make the interpolator.

    classify_models : list[str]
        The names of models or groups of models that are possible matches.
        This may have "\*"s to represent wild cards, hence multiple models will have
        all their data combined to make the interpolator.

    return_all_info : bool
        If True, instead of simply returning the strings specifying the closest
        scenario/model match, we return all scenario/model combinations in order of
        preference, along with the rms distance, quantifying the closeness.

    use_change_not_abs : bool
        If True, the code looks for the trend with the closest *derivatives* rather
        than the closest absolute value, i.e. closest trend allowing for an offset.
        This requires data from more than one time.

    Returns
    -------
    if return_all_info == False:

    (string, string)
        Strings specifying the model (first) and scenario (second) classifications
        that best match the data.

    if return_all_info == True:
    dict
        Maps the model and scenario classification strings to the measure of
        closeness.


    Raises
    ------
    ValueError
        Not all required timepoints are present in the database we crunched, we have
         `{dates we have}` but you passed in `{dates we need}`."
    """
    assert all(
        x in options_df.variables().values
        for x in [variable_follower] + variable_leaders
    ), "Not all required data is present in compared series"
    assert len(variable_leaders) == 1, "This is only calibrated to work with one leader"
    time_col = options_df.time_col
    assert (
        to_compare_df.time_col == time_col
    ), "The time column in the data to classify does not match the cruncher"
    times_needed = set(to_compare_df.data[time_col])
    if any(x not in options_df.data[time_col].values for x in times_needed):
        raise ValueError(
            "Not all required timepoints are present in the database we "
            "crunched, we have `{}` but you passed in {}.".format(
                list(set(options_df.data[time_col])),
                list(set(to_compare_df.data[time_col])),
            )
        )
    assert (
        len(times_needed) > 1 or use_change_not_abs == False
    ), "We need data from multiple times in order to calculate a difference."
    if to_compare_df.data.empty:
        print("The database being compared is empty")
        return None
    scen_model_rating = {}
    to_compare_db = _make_wide_db(to_compare_df)
    if use_change_not_abs:
        # Set all values to 0 at time 0 to remove any offset
        _remove_t0_from_wide_db(times_needed, to_compare_db)
    to_compare_db = to_compare_db.reset_index()
    for scenario in classify_scenarios:
        for model in classify_models:
            scenario_db = options_df.filter(
                scenario=scenario,
                model=model,
                variable=variable_leaders + [variable_follower],
            )
            if scenario_db.data.empty:
                scen_model_rating[model, scenario] = np.inf
                logger.warning(
                    "Data with scenario {} and model {} not found in data".format(
                        scenario, model
                    )
                )
                continue

            wide_db = _make_wide_db(scenario_db)
            if use_change_not_abs:
                # Set all values to 0 at time 0
                _remove_t0_from_wide_db(times_needed, wide_db)
            squared_dif = 0
            for leader in variable_leaders:
                all_interps = _make_interpolator(
                    variable_follower, leader, wide_db, time_col
                )
                for row in to_compare_db.iterrows():
                    squared_dif += (
                        row[1][variable_follower]
                        - all_interps[row[1][time_col]](row[1][leader])
                    ) ** 2
            scen_model_rating[model, scenario] = squared_dif
    ordered_scen = sorted(scen_model_rating.items(), key=lambda item: item[1])
    if return_all_info:
        return ordered_scen
    return ordered_scen[0][0]


def _remove_t0_from_wide_db(times_needed, _db):
    """
    This function finds the first set of values for each model and scenario and
    subtracts them from all values to remove the offset.
    """

    for model, scenario in set(
        zip(_db.index.get_level_values("model"), _db.index.get_level_values("scenario"))
    ):
        offset = _db.loc[model, scenario, min(times_needed)].copy().values.squeeze()
        for time in times_needed:
            _db.loc[model, scenario, time] = _db.loc[model, scenario, time] - offset


def _make_interpolator(variable_follower, variable_leader, wide_db, time_col):
    """
    Constructs a linear interpolator for variable_follower as a function of
    (one) variable_leader for each timestep in the data.
    """
    derived_relationships = {}
    for db_time, dbtdf in wide_db.groupby(time_col):
        xs = dbtdf[variable_leader].values.squeeze()
        ys = dbtdf[variable_follower].values.squeeze()
        if xs.shape != ys.shape:
            raise NotImplementedError(
                "Having more than one `variable_leaders` is not yet implemented"
            )
        if not xs.shape:
            # 0D-array, so we can return a single value
            xs = np.array([xs])
            ys = np.array([ys])
        # Ensure that any duplicates are replaced by their average value
        xs_pandas = pd.Series(xs)
        for x_dup in xs_pandas[xs_pandas.duplicated()]:
            inds = np.asarray(xs == x_dup).nonzero()[0]
            ys[inds[0]] = ys[inds].mean()
            xs = np.delete(xs, inds[1:])
            ys = np.delete(ys, inds[1:])
        xs, ys = map(np.array, zip(*sorted(zip(xs, ys))))
        if xs.shape == (1,):
            # If there is only one point, we must duplicate the data for interpolate
            xs = np.append(xs, xs)
            ys = np.append(ys, ys)
        derived_relationships[db_time] = scipy.interpolate.interp1d(
            xs, ys, bounds_error=False, fill_value=(ys[0], ys[-1]), assume_sorted=True
        )
    return derived_relationships


def _make_wide_db(use_db):
    """
    Converts an IamDataFrame into a pandas DataFrame that describes the timeseries
    of variables in index-labelled values.
    """
    idx = ["model", "scenario", use_db.time_col]
    assert (
        use_db.data.groupby(idx + ["variable"]).count()._get_values.max() <= 1
    ), "The table contains multiple entries with the same model and scenario"
    use_db = use_db.pivot_table(index=idx, columns="variable", aggfunc="sum")
    # make sure we don't have empty strings floating around (pyam bug?)
    use_db = use_db.applymap(lambda x: np.nan if isinstance(x, str) else x)
    use_db = use_db.dropna(axis=0)
    return use_db


def _get_unit_of_variable(df, variable, multiple_units="raise"):
    """
    Get the unit of a variable in ``df``

    Parameters
    ----------
    variable : str
        String to use to filter variables

    multiple_units : str
        If ``"raise"``, check that the variable only has one unit and raise an
        ``AssertionError`` if it has more than one unit.

    Returns
    -------
    list
        List of units for the variable

    Raises
    ------
    AssertionError
        ``multiple_units=="raise"`` and the filter results in more than one unit
    """
    units = df.filter(variable=variable).data["unit"].unique()
    if multiple_units == "raise":
        if len(units) > 1:
            raise AssertionError("`{}` has multiple units".format(variable))
        return units

    return units


def return_cases_which_consistently_split(
    df, aggregate, components, how_close=None, use_ar4_data=False
):
    """
    Returns model-scenario tuples which correctly split up the to_split into the various
    components. Components may contain wildcard "\*"s to match several variables.

    Parameters
    ----------
    df: :obj:`pyam.IamDataFrame`
        The input dataframe.

    aggregate : str
        Name of the variable that should split into the others

    components : list[str]
        List of the variable names whose sum should equal the to_split value (if
        expressed in common units).

    how_close : dict
        This is a dictionary of numpy.isclose options specifying how exact the match
        must be for the case to be included as passing. By default we specify a relative
        tolerance of 1% ('rtol': 1e-2). The syntax for this can be found in the numpy
        documentation.

    use_ar4_data : bool
        Determines whether the unit conversion takes place using GWP100 values from
        the UNFCCC AR5 (if false, default) or AR4 (if true).

    Returns
    -------
    list[(str, str, str)]
        List of consistent (Model name, scenario name, region name) tuples.
    """
    if not how_close:
        how_close = {"equal_nan": True, "rtol": 1e-02}
    valid_model_scenario = []
    df = convert_units_to_MtCO2_equiv(
        df.filter(variable=[aggregate] + components), use_ar4_data
    )
    combinations = df.data[["model", "scenario", "region"]].drop_duplicates()
    for ind in range(len(combinations)):
        model, scenario, region = combinations.iloc[ind]
        model_df = df.filter(model=model, scenario=scenario, region=region)
        # The following will often release a warning for empty data
        logging.getLogger("pyam.core").setLevel(logging.CRITICAL)
        to_split_df = model_df.filter(variable=aggregate)
        logging.getLogger("pyam.core").setLevel(logging.WARNING)
        if to_split_df.data.empty:
            continue
        sum_all = model_df.data.groupby(model_df.time_col).agg("sum")
        sum_to_split = to_split_df.data.groupby(model_df.time_col).agg("sum")
        if all(
            [
                np.isclose(
                    sum_all["value"].loc[time],
                    sum_to_split["value"].loc[time] * 2,
                    **how_close,
                )
                for time in sum_to_split.index
            ]
        ):
            valid_model_scenario.append((model, scenario, region))
    return valid_model_scenario


def convert_units_to_MtCO2_equiv(df, use_ar4_data=False):
    """
    Converts the units of gases reported in kt into Mt CO2 equivalent per year

    Uses GWP100 values from either (by default) AR5 or AR4 IPCC reports.

    Parameters
    ----------
    df : :obj:`pyam.IamDataFrame`
        The input dataframe whose units need to be converted.

    use_ar4_data : bool
        If true, use the AR4 GWP100 conversion figures, else use the AR5.

    Return
    ------
    :obj:`pyam.IamDataFrame`
        The input data with units converted.
    """
    # Check things need converting
    convert_to_str = "Mt CO2-equiv/yr"
    convert_to_str_clean = "Mt CO2/yr"
    if df["unit"].isin([convert_to_str, convert_to_str_clean]).all():
        return df

    context = "AR4GWP100" if use_ar4_data else "AR5GWP100"

    to_convert_df = df.copy()
    to_convert_var = to_convert_df.filter(
        unit=[convert_to_str, convert_to_str_clean], keep=False
    ).variables(True)
    to_convert_units = to_convert_var["unit"]
    to_convert_units_clean = {
        unit: unit.replace("-equiv", "").replace("equiv", "")
        for unit in to_convert_units
    }

    conversion_factors = {}
    not_found = []
    with _ur.context(context):
        for unit in to_convert_units:
            if unit in conversion_factors:
                continue

            clean_unit = to_convert_units_clean[unit]
            try:
                conversion_factors[unit] = (
                    _ur(clean_unit).to(convert_to_str_clean).magnitude
                )
            except DimensionalityError:
                raise ValueError(
                    "Cannot convert from {} (cleaned is: {}) "
                    "to {} (cleaned is: {})".format(
                        unit, clean_unit, convert_to_str, convert_to_str_clean
                    )
                )

    assert not not_found, "Not all units can be converted. We lack {}".format(not_found)

    for unit in to_convert_units:
        to_convert_df.convert_unit(
            {unit: [convert_to_str, conversion_factors[unit]]}, inplace=True
        )

    return to_convert_df


def download_or_load_sr15(filename, valid_model_ids="*"):
    """
    Load SR1.5 data, if it isn't there, download it

    Parameters
    ----------
    filename : str
        Filename in which to look for/save the data
    valid_model_ids : str
        Models to return from date

    Returns
    -------
    :obj: `pyam.IamDataFrame`
        The loaded data
    """
    if not os.path.isfile(filename):
        get_sr15_scenarios(filename, valid_model_ids)
    return pyam.IamDataFrame(filename).filter(model=valid_model_ids)


def get_sr15_scenarios(output_file, valid_model_ids):
    """
    Collects world-level data from the IIASA database for the named models and saves them to a given location.

    Parameters
    ----------
    output_file : str
        File name and location for data to be saved

    valid_model_ids : list[str]
        Names of models that are to be fetched.
    """
    conn = pyam.iiasa.Connection("IXSE_SR15")
    variables_to_fetch = ["Emissions*"]
    for model in valid_model_ids:
        print("Fetching data for {}".format(model))
        for variable in variables_to_fetch:
            print("Fetching {}".format(variable))
            var_df = conn.query(model=model, variable=variable, region="World")
            try:
                df.append(var_df, inplace=True)
            except NameError:
                df = pyam.IamDataFrame(var_df)

    print("Writing to {}".format(output_file))
    df.to_csv(output_file)


def _adjust_time_style_to_match(in_df, target_df):
    if in_df.time_col != target_df.time_col:
        in_df = in_df.timeseries()
        if target_df.time_col == "time":
            target_df_year_map = {v.year: v for v in target_df.timeseries().columns}
            in_df.columns = in_df.columns.map(
                lambda x: target_df_year_map[x]
                if x in target_df_year_map
                else dt.datetime(x, 1, 1)
            )
        else:
            in_df.columns = in_df.columns.map(lambda x: x.year)
        return pyam.IamDataFrame(in_df)

    return in_df


def _construct_consistent_values(aggregate_name, components, db_to_generate):
    """
    Calculates the sum of the components and creates an IamDataFrame with this
    value under variable type `aggregate_name`.

    Parameters
    ----------
    aggregate_name : str
        The name of the aggregate variable.

    components : [str]
        List of the names of the variables to be summed.

    db_to_generate : :obj:`pyam.IamDataFrame`
        Input data from which to construct consistent values.

    Returns
    -------
    :obj:`pyam.IamDataFrame`
        Consistently calculated aggregate data.
    """
    assert (
        aggregate_name not in db_to_generate.variables().values
    ), "We already have a variable of this name"
    relevant_db = db_to_generate.filter(variable=components)
    units = relevant_db.data["unit"].drop_duplicates().sort_values()
    unit_equivs = units.map(lambda x: x.replace("-equiv", "")).drop_duplicates()
    if len(unit_equivs) == 0:
        raise ValueError(
            "Attempting to construct a consistent {} but none of the components "
            "present".format(aggregate_name)
        )
    elif len(unit_equivs) > 1:
        raise ValueError(
            "Too many units found to make a consistent {}".format(aggregate_name)
        )
    use = (
        relevant_db.data.groupby(["model", "scenario", "region", relevant_db.time_col])
        .agg("sum")
        .reset_index()
    )
    # Units are sorted in alphabetical order so we choose the first to get -equiv
    use["unit"] = units.iloc[0]
    use["variable"] = aggregate_name
    return pyam.IamDataFrame(use)



"""
Silicone's custom statistical operations.
"""
import os

import numpy as np
import pandas as pd


def rolling_window_find_quantiles(
    xs, ys, quantiles, nwindows=10, decay_length_factor=1
):
    """
    Perform quantile analysis in the y-direction for x-weighted data.

    Divides the x-axis into nwindows of equal length and weights data by how close they
    are to the center of these boxes. Then returns the quantiles of this weighted data.
    Quantiles are defined so that the values returned are always equal to a y-value in
    the data - there is no interpolation. Extremal points are given their full
    weighting, meaning this will not agree with the np.quantiles under uniform weighting
    (which effectively gives 0 weight to min and max values)

    The weighting of a point at :math:`x` for a window centered at :math:`x_0` is:

    .. math::

        w = \\frac{1}{1 + \\left (\\frac{x - x_0}{ \\text{box\\_length} \\times \\text{decay\\_length\\_factor} } \\right)^2}

    Parameters
    ----------
    xs : np.ndarray, :obj:`pd.Series`
        The x co-ordinates to use in regression.

    xs : np.ndarray, :obj:`pd.Series`
        The x co-ordinates to use in regression.

    quantiles : list-like
        The quantiles to calculate in each window

    nwindows : positive int
        How many points to evaluate between x_max and x_min.

    decay_length_factor : float
        gives the distance over which the weighting of the values falls to 1/4,
        relative to half the distance between boxes. Defaults to 1. Formula is
        :math:`w = \\left ( 1 + \\left( \\frac{\\text{distance}}{\\text{box\\_length} \\times \\text{decay\\_length\\_factor}} \\right)^2 \\right)^{-1}`.

    Returns
    -------
    :obj:`pd.DataFrame`
        Quantile values at the window centres.
    """
    assert xs.size == ys.size
    if xs.size == 1:
        return pd.DataFrame(index=[xs[0]] * nwindows, columns=quantiles, data=ys[0])
    step = (max(xs) - min(xs)) / (nwindows + 1)
    decay_length = step / 2 * decay_length_factor
    # We re-form the arrays in case they were pandas series with integer labels that
    # would mess up the sorting.
    sort_order = np.argsort(ys)
    ys = ys[sort_order]
    xs = xs[sort_order]
    if max(xs) == min(xs):
        # We must prevent singularity behaviour if all the points have the same x.
        box_centers = np.array([xs[0]])
        decay_length = 1
    else:
        # We want to include the max x point, but not any point above it.
        # The 0.99 factor prevents rounding error inclusion.
        box_centers = np.arange(min(xs), max(xs) + step * 0.99, step)
    quantmatrix = pd.DataFrame(index=box_centers, columns=quantiles)
    for ind in range(box_centers.size):
        weights = 1.0 / (1.0 + ((xs - box_centers[ind]) / decay_length) ** 2)
        weights /= sum(weights)
        cumsum_weights = np.cumsum(weights)
        for i_quantile in range(quantiles.__len__()):
            quantmatrix.iloc[ind, i_quantile] = min(
                ys[cumsum_weights >= quantiles[i_quantile]]
            )
    return quantmatrix


def calc_all_emissions_correlations(emms_df, years, output_dir):
    """
    Save csv files of the correlation coefficients and the rank correlation coefficients between emissions at specified locations.

    This function includes all undivided emissions (i.e. results recorded as `Emissions|X`) and CO2 emissions split once (i.e. `Emissions|CO2|X`). It does not include Kyoto gases. It will also save the average absolute value of the coefficients.

    Files created:

    - "variable_counts.csv" : the number of scenario/model pairs where the emissions data occurs.

    - "gases_correlation_{year}.csv" : The Pearson's correlation between gases emissions in a given year.

    - "gases_rank_correlation_{year}.csv" : The Spearman's rank correlation between gases in a given year

    - "time_av_absolute_correlation_{}_to_{}.csv" : The magnitude of the Pearson's correlation between emissions, averaged over the years requested.

    - "time_av_absolute_rank_correlation_{}_to_{}.csv" : The magnitude of the Spearman's rank correlation between emissions, averaged over the years requested.

    Parameters
    ----------
    emms_df : :obj:`pyam.IamDataFrame`
        The database to search for correlations between named values

    output_dir : str
        The folder location to save the files.

    years : list[int]
        The years upon which to calculate correlations.
    """
    assert len(emms_df.regions()) == 1, "Calculation is for only one region"
    # Obtain the list of gases to examine
    df_gases = (
        emms_df.filter(level=1)
        .filter(variable="Emissions|*")
        .filter(variable="Emissions|Kyoto*", keep=False)
        .append(emms_df.filter(level=2).filter(variable="Emissions|CO2*"))
        .variables(True)
        .set_index("variable")
    )
    all_correlations_df = pd.DataFrame(
        index=df_gases.index, columns=df_gases.index, data=0
    )
    all_rank_corr_df = pd.DataFrame(
        index=df_gases.index, columns=df_gases.index, data=0
    )

    # Calculate the total amount of data
    var_count_file = "variable_counts.csv"
    var_count = pd.Series(index=df_gases.index, dtype=int)
    for var in df_gases.index:
        var_db = emms_df.filter(variable=var)
        var_count[var] = len(var_db.timeseries())
    var_save_loc = os.path.join(output_dir, var_count_file)
    var_count.to_csv(var_save_loc)
    print("Counted the number of each variable and saved to ".format(var_save_loc))

    for year_of_interest in years:
        # Initialise the tables to hold all parameters between runs
        correlations_df = pd.DataFrame(index=df_gases.index, columns=df_gases.index)
        rank_corr_df = pd.DataFrame(index=df_gases.index, columns=df_gases.index)
        # Check that the list has only one entry for each gas
        assert not any(df_gases.index.duplicated()), "Index contains duplicated entries"
        formatted_df = emms_df.filter(
            variable=df_gases.index, year=year_of_interest
        ).pivot_table(
            ["year", "model", "scenario", "region"], ["variable"], aggfunc="mean"
        )
        formatted_df.replace(r"^\s*$", np.nan, regex=True, inplace=True)
        for x_gas_ind in range(df_gases.count()[0]):
            x_gas = df_gases.index[x_gas_ind]
            for y_gas_ind in range(x_gas_ind + 1, df_gases.count()[0]):
                y_gas = df_gases.index[y_gas_ind]
                # Calculate the correlations. This requires removing NAs
                correlations_df.at[y_gas, x_gas] = formatted_df.corr("pearson").loc[
                    x_gas, y_gas
                ]
                rank_corr_df.at[y_gas, x_gas] = formatted_df.corr("spearman").loc[
                    x_gas, y_gas
                ]
                all_correlations_df.loc[y_gas, x_gas] = all_correlations_df.at[
                    y_gas, x_gas
                ] + abs(correlations_df.loc[y_gas, x_gas]) / len(years)
                all_rank_corr_df.loc[y_gas, x_gas] = all_rank_corr_df.at[
                    y_gas, x_gas
                ] + abs(rank_corr_df.at[y_gas, x_gas]) / len(years)
                # the other parts follow by symmetry
                correlations_df.at[x_gas, y_gas] = correlations_df.at[y_gas, x_gas]
                rank_corr_df.at[x_gas, y_gas] = rank_corr_df.at[y_gas, x_gas]
                all_correlations_df.loc[x_gas, y_gas] = all_correlations_df.at[
                    y_gas, x_gas
                ]
                all_rank_corr_df.loc[x_gas, y_gas] = all_rank_corr_df.at[y_gas, x_gas]
            print("Finished x_gas {} in year {}.".format(x_gas, year_of_interest))
        if output_dir is not None:
            correlations_df.to_csv(
                os.path.join(
                    output_dir, "gases_correlation_{}.csv".format(year_of_interest)
                )
            )
            rank_corr_df.to_csv(
                os.path.join(
                    output_dir, "gases_rank_correlation_{}.csv".format(year_of_interest)
                )
            )
    for gas in df_gases.index:
        all_correlations_df.loc[gas, gas] = np.nan
        all_rank_corr_df.loc[gas, gas] = np.nan
    if output_dir is not None:
        all_correlations_df.to_csv(
            os.path.join(
                output_dir,
                "time_av_absolute_correlation_{}_to_{}.csv".format(
                    min(years), max(years)
                ),
            )
        )
        all_rank_corr_df.to_csv(
            os.path.join(
                output_dir,
                "time_av_absolute_rank_correlation_{}_to_{}.csv".format(
                    min(years), max(years)
                ),
            )
        )



import logging
import re
import warnings

import pyam
import tqdm

from silicone.database_crunchers import (
    DatabaseCruncherConstantRatio,
    DatabaseCruncherQuantileRollingWindows,
)


"""
Infills all required data for MAGICC and FAIR emulators with minimal configuration 
"""


def infill_all_required_variables(
    to_fill,
    database,
    variable_leaders,
    required_variables_list=None,
    cruncher=DatabaseCruncherQuantileRollingWindows,
    output_timesteps=None,
    infilled_data_prefix=None,
    to_fill_old_prefix=None,
    check_data_returned=False,
):
    """
    This is function designed to infill all required data given a minimal amount of
    input.

    Parameters
    ----------
    to_fill : :obj:`pyam.IamDataFrame`
        The dataframe which is to be infilled

    database: :obj:`pyam.IamDataFrame`
        The dataframe containing all information to be used in the infilling process.

    variable_leaders: list[str]
        The name of the variable(s) found in to_fill which should be used to determine
        the values of the other variables. For most infillers (including the default)
        this list must contain only one entry. E.g. ["Emissions|CO2"]

    required_variables_list: list[str]
        The list of variables to infill. Each will be done separately. The default
        behaviour (None option) will result in this being filled with the complete list
        of required emissions.

    cruncher : :class:
        The class of cruncher to use to compute the infilled values. Defaults to
        DatabaseCruncherQuantileRollingWindows, which uses the median value of a rolling
        window. See the cruncher documentation for more details.

    output_timesteps : list[int or datetime]
        List of times at which to return infilled values. Will interpolate values in
        between known data, but will not extend beyond the range of data provided.

    infilled_data_prefix : str
        A string that should be prefixed on all the variable names of the results
        returned. Used to distinguish returned values from those input.

    to_fill_old_prefix : str
        Any string already found at the beginning of the variables names of the input
        `to_fill` dataframe. This will be removed before comparing the variable names
        with `database`.

    check_data_returned : bool
        If true, we perform checks that all desired data has been returned. Potential
        reasons for failing this include requesting results at times outside our input
        time range, as well as code bugs.

    Returns
    -------
    :obj:`pyam.IamDataFrame`
        The infilled dataframe (including input data) at requested times. All variables
        now begin with infilled_data_prefix instead of to_fill_old_prefix.
    """
    # Use default arguments for unfilled options:
    if output_timesteps is None:
        output_timesteps = [2015] + list(range(2020, 2101, 10))
    if required_variables_list is None:
        required_variables_list = [
            "Emissions|BC",
            "Emissions|PFC|CF4",
            "Emissions|PFC|C2F6",
            "Emissions|PFC|C6F14",
            "Emissions|CH4",
            "Emissions|CO2|AFOLU",
            "Emissions|CO2|Energy and Industrial Processes",
            "Emissions|CO",
            "Emissions|HFC|HFC134a",
            "Emissions|HFC|HFC143a",
            "Emissions|HFC|HFC227ea",
            "Emissions|HFC|HFC23",
            "Emissions|HFC|HFC32",
            "Emissions|HFC|HFC43-10",
            "Emissions|HFC|HFC245ca",
            "Emissions|HFC|HFC125",
            "Emissions|N2O",
            "Emissions|NH3",
            "Emissions|NOx",
            "Emissions|OC",
            "Emissions|SF6",
            "Emissions|Sulfur",
            "Emissions|VOC",
        ]

    # Check that the input is valid
    if to_fill_old_prefix:
        if any(
            to_fill.data["variable"].map(lambda x: x[: len(to_fill_old_prefix)])
            != to_fill_old_prefix
        ):
            raise ValueError("Not all of the data begins with the expected prefix")
        to_fill.data["variable"] = to_fill.data["variable"].str.replace(
            re.escape(to_fill_old_prefix + "|"), ""
        )
    if infilled_data_prefix:
        if any(
            to_fill.data["variable"].map(lambda x: x[: len(infilled_data_prefix)])
            == infilled_data_prefix
        ):
            raise ValueError(
                "This data already contains values with the expected final "
                "prefix. This suggests that some of it has already been infilled."
            )
    assert len(to_fill.regions()) == 1, "There are {} regions in the data.".format(
        len(to_fill.regions())
    )
    assert len(database.regions()) == 1
    assert (
        to_fill.data["region"].iloc[0] == database.data["region"].iloc[0]
    ), "The cruncher data and the infilled data have different regions."
    # Perform any interpolations required here
    to_fill_orig = to_fill.copy()
    timecol = database.time_col
    assert timecol == to_fill.time_col
    df_times_missing = database.timeseries().isna().sum() > 0
    to_fill_times_missing = to_fill.timeseries().isna().sum() > 0
    for time in output_timesteps:
        if time not in database[timecol].tolist() or df_times_missing[time]:
            database.interpolate(time)
        if time not in to_fill[timecol].tolist() or to_fill_times_missing[time]:
            to_fill.interpolate(time)
    # Nans in additional columns break pyam, so we overwrite them
    database.data[database.extra_cols] = database.data[database.extra_cols].fillna(0)
    to_fill.data[to_fill.extra_cols] = to_fill.data[to_fill.extra_cols].fillna(0)
    # Filter for desired times
    if timecol == "year":
        database = database.filter(year=output_timesteps)
        to_fill = to_fill.filter(year=output_timesteps)
    else:
        database = database.filter(time=output_timesteps)
        to_fill = to_fill.filter(time=output_timesteps)
    # Infill unavailable data
    assert not database.data.isnull().any().any()
    assert not to_fill.data.isnull().any().any()
    unavailable_variables = [
        variab
        for variab in required_variables_list
        if variab not in database.variables().values
    ]
    if unavailable_variables:
        warnings.warn(UserWarning("No data for {}".format(unavailable_variables)))
        # Infill the required variables with 0s.
        kwarg_dict = {"ratio": 0, "units": "Mt CO2-equiv/yr"}
        to_fill = _perform_crunch_and_check(
            unavailable_variables,
            variable_leaders,
            to_fill,
            database,
            DatabaseCruncherConstantRatio,
            output_timesteps,
            to_fill_orig,
            check_data_returned=False,
            **kwarg_dict,
        )
    available_variables = [
        variab
        for variab in required_variables_list
        if variab not in unavailable_variables
    ]
    if available_variables:
        to_fill = _perform_crunch_and_check(
            available_variables,
            variable_leaders,
            to_fill,
            database,
            cruncher,
            output_timesteps,
            to_fill_orig,
            check_data_returned=check_data_returned,
        )
    if infilled_data_prefix:
        to_fill.data["variable"] = infilled_data_prefix + "|" + to_fill.data["variable"]
    return to_fill


def _perform_crunch_and_check(
    required_variables,
    leaders,
    to_fill,
    df,
    type_of_cruncher,
    output_timesteps,
    to_fill_orig,
    check_data_returned=False,
    **kwargs,
):
    """
        Takes a list of scenarios to infill and infills them according to the options
        presented.

        Parameters
        ----------
        required_variables : list[str]
            The variable names to infill

        leaders : list[str]
            The leaders to guide the infilling

        to_fill : IamDataFrame
            The data frame to infill

        df : IamDataFrame
            The data frame to base the infilling on

        type_of_cruncher : :obj: silicone cruncher
            the silicone package cruncher class to use for the infilling

        output_timesteps : list[int or datetime]
            When there should be data returned. Time-based interpolation will occur if
            this is more frequent than the data allows, data will be filtered out if
            there is additional time information.

        to_fill_orig : IamDataFrame
            The original, unfiltered and unaltered data input. We use this for
            performing checks.

        kwargs : Dict
            Any key word arguments to include in the cruncher calculation

        Returns
        -------
        :obj:IamDataFrame
            The infilled dataframe
        """
    if (
        not all(x in df.variables().values for x in required_variables)
        and type_of_cruncher != DatabaseCruncherConstantRatio
    ):
        not_present = [x for x in required_variables if x not in df.variables().values]
        raise Warning("Missing some requested variables: {}".format(not_present))
    cruncher = type_of_cruncher(df)
    for req_var in tqdm.tqdm(required_variables, desc="Filling required variables"):
        interpolated = _infill_variable(cruncher, req_var, leaders, to_fill, **kwargs)
        if interpolated:
            to_fill = to_fill.append(interpolated)
    # Optionally check we have added all the required data
    if not check_data_returned:
        return to_fill
    for _, (model, scenario) in (
        to_fill[["model", "scenario"]].drop_duplicates().iterrows()
    ):
        msdf = to_fill.filter(model=model, scenario=scenario)
        for v in required_variables:
            msvdf = msdf.filter(variable=v)
            msvdf_data = msvdf.data
            assert not msvdf_data.isnull().any().any()
            assert not msvdf_data.empty
            assert all(
                [y in msvdf_data[df.time_col].values for y in output_timesteps]
            ), "We do not have data for all required timesteps"

    # Check no data was overwritten by accident
    for model in tqdm.tqdm(
        to_fill_orig.models(), desc="Consistency with original model data checks"
    ):
        mdf = to_fill_orig.filter(model=model, variable=leaders + required_variables)
        for scenario in mdf.scenarios():
            msdf = mdf.filter(scenario=scenario)
            msdf_filled = to_fill.filter(
                model=model, scenario=scenario, variable=msdf["variable"].unique()
            )

            common_times = set(msdf_filled[msdf.time_col]).intersection(
                msdf[msdf.time_col]
            )
            if common_times:
                if msdf.time_col == "year":
                    msdf = msdf.filter(year=list(common_times))
                    msdf_filled = msdf_filled.filter(year=list(common_times))
                else:
                    msdf = msdf.filter(time=list(common_times))
                    msdf_filled = msdf_filled.filter(time=list(common_times))
                assert pyam.compare(msdf, msdf_filled).empty
    return to_fill


def _infill_variable(cruncher_i, req_variable, leader_i, to_fill_i, **kwargs):
    """
    A function used to iterate the actual crunching if the data doesn't already
    exist.
    Parameters
    ----------
    cruncher_i : :obj: silicone cruncher
        the initiated silicone cruncher to use for the infilling

    req_variable : str
        The follower variable to infill.

    leader_i : list[str]
        The leader variable to guide the infilling.

    to_fill_i : IamDataFrame
        The dataframe to infill.

    kwargs : Dict
        Any key word arguments to include in the cruncher calculation

    Returns
    -------
    :obj:IamDataFrame
        The infilled component of the dataframe (or None if no infilling done)
    """
    filler = cruncher_i.derive_relationship(req_variable, leader_i, **kwargs)
    # only fill for scenarios who don't have that variable
    # quieten logging about empty data frame as it doesn't matter here
    logging.getLogger("pyam.core").setLevel(logging.CRITICAL)
    not_to_fill = to_fill_i.filter(variable=req_variable)

    to_fill_var = to_fill_i.copy()
    if not not_to_fill.data.empty:
        for (model, scenario), _ in not_to_fill.data.groupby(["model", "scenario"]):
            to_fill_var = to_fill_var.filter(model=model, scenario=scenario, keep=False)
    if not to_fill_var.data.empty:
        interpolated = filler(to_fill_var)
        return interpolated
    logging.getLogger("pyam.core").setLevel(logging.WARNING)
    return None



"""
Module for the database cruncher which uses the 'rolling windows' technique.
"""

import logging

import numpy as np
import pandas as pd
import scipy.interpolate
from pyam import IamDataFrame

from ..utils import _get_unit_of_variable
from .base import _DatabaseCruncher

logger = logging.getLogger(__name__)


class DatabaseCruncherQuantileRollingWindows(_DatabaseCruncher):
    """
    Database cruncher which uses the 'rolling windows' technique.

    This cruncher derives the relationship between two variables by performing
    quantile calculations between the follower timeseries and the lead timeseries.
    These calculations are performed at each timestep in the timeseries, independent
    of the other timesteps.

    For each timestep, the lead timeseries axis is divided into multiple evenly spaced
    windows (to date this is only tested on 1:1 relationships but may work with more
    than one lead timeseries). In each window, every data point in the database is
    included. However, the data points receive a weight given by

    .. math::

        w(x, x_{\\text{window}}) = \\frac{1}{1 + (d_n)^2}

    where :math:`w` is the weight and :math:`d_n` is the normalised distance between
    the centre of the window and the data point's position on the lead timeseries axis.

    :math:`d_n` is calculated as

    .. math::

        d_n = \\frac{x - x_{\\text{window}}}{f \\times (\\frac{b}{2})}

    where :math:`x` is the position of the data point on the lead timeseries axis,
    :math:`x_{\\text{window}}` is the position of the centre of the window on the lead
    timeseries axis, :math:`b` is the distance between window centres and :math:`f` is
    a decay factor which controls how much less points away from :math:`x_{\\text{window}}`
    are weighted. If :math:`f=1` then a point which is halfway between
    window centres receives a weighting of :math:`1/2`. Lowering the value of
    :math:`f` cause points further from the window centre to receive less weight.

    With these weightings, the desired quantile of the data is then calculated. This
    calculation is done by sorting the data, and then choosing the first data point
    from the database which sits above or equal to the given quantile, with each data
    point's contribution to the quantile calculation being weighted by its weight. As
    a result, this cruncher limits itself to using data within the distribution and
    will not make any assumptions about the shape of the distribution.

    If the option ``use_ratio`` is set to ``True``, instead of returning the absolute
    value of the follow at this quantile, we return the quantile of the ratio between
    the lead and follow data in the database, multiplied by the actual lead value of the
    database being infilled.

    By varying the quantile, this cruncher can provide ranges of the relationship
    between different variables. For example, it can provide the 90th percentile (i.e.
    high end) of the relationship between e.g. ``Emissions|CH4`` and ``Emissions|CO2``
    or the 50th percentile (i.e. median) or any other arbitrary percentile/quantile
    choice.
    """

    def derive_relationship(
        self,
        variable_follower,
        variable_leaders,
        quantile=0.5,
        nwindows=10,
        decay_length_factor=1,
        use_ratio=False,
    ):
        """
        Derive the relationship between two variables from the database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CH4"``).

        variable_leaders : list[str]
            The variable(s) we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``).

        quantile : float
            The quantile to return in each window.

        nboxes : int
            The number of windows to use when calculating the relationship between the
            follower and lead gases.

        decay_length_factor : float
            Parameter which controls how strongly points away from the window's centre
            should be weighted compared to points at the centre. Larger values give
            points further away increasingly less weight, smaller values give points
            further away increasingly more weight.

        use_ratio : bool
            If false, we use the quantile value of the weighted mean absolute value. If
            true, we find the quantile weighted mean ratio between lead and follow,
            then multiply the ratio by the input value.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.

        Raises
        ------
        ValueError
            There is no data for ``variable_leaders`` or ``variable_follower`` in the
            database.

        ValueError
            ``quantile`` is not between 0 and 1.

        ValueError
            ``nwindows`` is not equivalent to an integer.

        ValueError
            ``decay_length_factor`` is 0.
        """
        self._check_follower_and_leader_in_db(variable_follower, variable_leaders)

        if not (0 <= quantile <= 1):
            error_msg = "Invalid quantile ({}), it must be in [0, 1]".format(quantile)
            raise ValueError(error_msg)

        if int(nwindows) != nwindows:
            error_msg = "Invalid nwindows ({}), it must be an integer".format(nwindows)
            raise ValueError(error_msg)
        nwindows = int(nwindows)

        if np.equal(decay_length_factor, 0):
            raise ValueError("decay_length_factor must not be zero")

        data_leader_unit = _get_unit_of_variable(self._db, variable_leaders)[0]
        data_follower_unit = _get_unit_of_variable(self._db, variable_follower)[0]
        db_time_col = self._db.time_col

        columns = "variable"
        idx = list(set(self._db.data.columns) - {columns, "value", "unit"})
        wide_db = self._db.filter(
            variable=[variable_follower] + variable_leaders
        ).pivot_table(index=idx, columns=columns, aggfunc="sum")
        # make sure we don't have empty strings floating around (pyam bug?)
        wide_db = wide_db.applymap(lambda x: np.nan if isinstance(x, str) else x)
        wide_db = wide_db.dropna(axis=0)

        derived_relationships = {}
        for db_time, dbtdf in wide_db.groupby(db_time_col):
            xs = dbtdf[variable_leaders].values.squeeze()
            ys = dbtdf[variable_follower].values.squeeze()
            if xs.shape != ys.shape:
                raise NotImplementedError(
                    "Having more than one `variable_leaders` is not yet implemented"
                )

            if not xs.shape:
                # 0D-array, make 1D
                xs = np.array([xs])
                ys = np.array([ys])

            step = (max(xs) - min(xs)) / nwindows
            decay_length = step / 2 * decay_length_factor
            if use_ratio:
                # We want the ratio between x and y, not the actual values of y.
                ys = ys / xs
                if np.isnan(ys).any():
                    logging.warning(
                        "Undefined values of ratio appear in the quantiles when "
                        "infilling {}, setting some values to 0 (this may not affect "
                        "results).".format(variable_follower)
                    )
                    ys[np.isnan(ys)] = 0
            sort_order = np.argsort(ys)
            ys = ys[sort_order]
            xs = xs[sort_order]
            if max(xs) == min(xs):
                # We must prevent singularity behaviour if all the points are at the
                # same x value.
                cumsum_weights = np.array([(1 + x) / len(ys) for x in range(len(ys))])

                def same_x_val_workaround(
                    _, ys=ys, cumsum_weights=cumsum_weights, quantile=quantile
                ):
                    return min(ys[cumsum_weights >= quantile])

                derived_relationships[db_time] = same_x_val_workaround
            else:
                # We want to include the max x point, but not any point above it.
                # The 0.99 factor prevents rounding error inclusion.
                window_centers = np.arange(min(xs), max(xs) + step * 0.99, step)

                db_time_table = pd.DataFrame(
                    index=pd.MultiIndex.from_arrays(
                        ([db_time], [quantile]), names=["db_time", "quantile"]
                    ),
                    columns=window_centers,
                )
                db_time_table.columns.name = "window_centers"

                for window_center in window_centers:
                    weights = 1.0 / (1.0 + ((xs - window_center) / decay_length) ** 2)
                    weights /= sum(weights)
                    # We want to calculate the weights at the midpoint of step
                    # corresponding to the y-value.
                    cumsum_weights = np.cumsum(weights)
                    db_time_table.loc[(db_time, quantile), window_center] = min(
                        ys[cumsum_weights >= quantile]
                    )

                derived_relationships[db_time] = scipy.interpolate.interp1d(
                    db_time_table.columns.values.squeeze(),
                    db_time_table.loc[(db_time, quantile), :].values.squeeze(),
                    bounds_error=False,
                    fill_value=(
                        db_time_table.loc[(db_time, quantile)].iloc[0],
                        db_time_table.loc[(db_time, quantile)].iloc[-1],
                    ),
                )

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`DatabaseCruncherQuantileRollingWindows`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                The key db_times for filling are not in ``in_iamdf``.
            """
            if db_time_col != in_iamdf.time_col:
                raise ValueError(
                    "`in_iamdf` time column must be the same as the time column used "
                    "to generate this filler function (`{}`)".format(db_time_col)
                )

            var_units = _get_unit_of_variable(in_iamdf, variable_leaders)
            if var_units.size == 0:
                raise ValueError(
                    "There is no data for {} so it cannot be infilled".format(
                        variable_leaders
                    )
                )
            var_units = var_units[0]

            if var_units != data_leader_unit:
                raise ValueError(
                    "Units of lead variable is meant to be `{}`, found `{}`".format(
                        data_leader_unit, var_units
                    )
                )

            # check whether we have all the required timepoints or not
            have_all_timepoints = all(
                [c in derived_relationships for c in in_iamdf.timeseries()]
            )

            if not have_all_timepoints:
                raise ValueError(
                    "Not all required timepoints are present in the database we "
                    "crunched, we crunched \n\t`{}`\nbut you passed in \n\t{}".format(
                        list(derived_relationships.keys()),
                        in_iamdf.timeseries().columns.tolist(),
                    )
                )

            # do infilling here
            infilled_ts = in_iamdf.filter(variable=variable_leaders).timeseries()
            for col in infilled_ts:
                if use_ratio:
                    infilled_ts[col] = (
                        derived_relationships[col](infilled_ts[col]) * infilled_ts[col]
                    )
                else:
                    infilled_ts[col] = derived_relationships[col](infilled_ts[col])

            infilled_ts = infilled_ts.reset_index()
            infilled_ts["variable"] = variable_follower
            infilled_ts["unit"] = data_follower_unit

            return IamDataFrame(infilled_ts)

        return filler



"""
Multiple infillers

Multiple infillers provide easy-to-use infiller options for the most common use-cases.
"""

from .decompose_collection_with_time_dep_ratio import (  # noqa: F401
    DecomposeCollectionTimeDepRatio,
)
from .infill_all_required_emissions_for_openscm import (  # noqa: F401
    infill_all_required_variables,
)
from .infill_composite_values import infill_composite_values  # noqa: F401



"""
Database crunchers.

The classes within this module can be used to crunch a database of scenarios. Each
'Cruncher' has methods which return functions which can then be used to infill
emissions detail (i.e. calculate 'follower' timeseries) based on 'lead' emissions
timeseries.
"""

from .constant_ratio import DatabaseCruncherConstantRatio  # noqa: F401
from .interpolate_specified_scenarios_and_models import (  # noqa: F401
    DatabaseCruncherScenarioAndModelSpecificInterpolate,
)
from .lead_gas import DatabaseCruncherLeadGas  # noqa: F401
from .linear_interpolation import DatabaseCruncherLinearInterpolation  # noqa: F401
from .quantile_rolling_windows import (  # noqa: F401
    DatabaseCruncherQuantileRollingWindows,
)
from .rms_closest import DatabaseCruncherRMSClosest  # noqa: F401
from .time_dep_ratio import DatabaseCruncherTimeDepRatio  # noqa: F401



"""
Module for the database cruncher which uses the 'lead gas' technique.
"""
import warnings

import numpy as np
from pyam import IamDataFrame

from .base import _DatabaseCruncher


class DatabaseCruncherLeadGas(_DatabaseCruncher):
    """
    Database cruncher which uses the 'lead gas' technique.

    This cruncher derives the relationship between two variables by simply assuming
    that the follower timeseries is equal to the lead timeseries multiplied by a
    scaling factor. The scaling factor is derived by calculating the ratio of the
    follower variable to the lead variable in the latest year in which the follower
    variable is available in the database. Additionally, since
    the derived relationship only depends on a single point in the database, no
    regressions or other calculations are performed.

    Once the relationship is derived, the 'filler' function will infill following:

    .. math::
        E_f(t) = R * E_l(t)

    where :math:`E_f(t)` is emissions of the follower variable and :math:`E_l(t)` is
    emissions of the lead variable, both in the infillee database.

    :math:`R` is the scaling factor, calculated as

    .. math::
        R = \\frac{ E_f(t_{\\text{last}}) }{ e_l(t_{\\text{last}}) }

    where :math:`t_{\\text{last}}` is the average of all values of the follower gas at the latest time it
    appears in the database, and the lower case :math:`e` represents the infiller
    database.
    """

    def derive_relationship(self, variable_follower, variable_leaders):
        """
        Derive the relationship between two variables from the database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|C5F12"``).

        variable_leaders : list[str]
            The variable we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``). Note that the 'lead
            gas' methodology gives the same result, independent of the value of
            ``variable_leaders`` in the database.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.

        Raises
        ------
        ValueError
            ``variable_leaders`` contains more than one variable.

        ValueError
            There is no data for ``variable_leaders`` or ``variable_follower`` in the
            database.

        """
        iamdf_follower = self._get_iamdf_follower(variable_follower, variable_leaders)
        data_follower = iamdf_follower.data

        data_follower_time_col = iamdf_follower.time_col
        data_follower_key_timepoint = max(data_follower[data_follower_time_col])
        key_timepoint_filter = {data_follower_time_col: [data_follower_key_timepoint]}
        data_follower_key_year_val = np.nanmean(
            iamdf_follower.filter(**key_timepoint_filter)["value"].values
        )
        data_follower_unit = data_follower["unit"].values[0]

        if data_follower_time_col == "time":
            data_follower_key_timepoint = data_follower_key_timepoint.to_pydatetime()

        def filler(in_iamdf, interpolate=False):
            """
            Filler function derived from :obj:`DatabaseCruncherLeadGas`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            interpolate : bool
                If the key year for filling is not in ``in_iamdf``, should a value be
                interpolated?

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                The key year for filling is not in ``in_iamdf`` and ``interpolate is
                False``.
            """
            lead_var = in_iamdf.filter(variable=variable_leaders)

            if data_follower_time_col != in_iamdf.time_col:
                raise ValueError(
                    "`in_iamdf` time column must be the same as the time column used "
                    "to generate this filler function (`{}`)".format(
                        data_follower_time_col
                    )
                )

            def get_values_in_key_timepoint(idf):
                # filter warning about empty data frame as we handle it ourselves
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    return idf.filter(**key_timepoint_filter)

            lead_var_val_in_key_timepoint = get_values_in_key_timepoint(lead_var)

            if lead_var_val_in_key_timepoint.data.empty:
                if not interpolate:
                    error_msg = (
                        "Required downscaling timepoint ({}) is not in the data for "
                        "the lead gas ({})".format(
                            data_follower_key_timepoint, variable_leaders[0]
                        )
                    )
                    raise ValueError(error_msg)
                else:
                    # TODO: make interpolate method of IamDataFrame datetime friendly
                    lead_var_interp = lead_var.timeseries()
                    lead_var_interp[data_follower_key_timepoint] = np.nan
                    lead_var_interp = lead_var_interp.reindex(
                        sorted(lead_var_interp.columns), axis=1
                    )
                    lead_var_interp = IamDataFrame(
                        lead_var_interp.interpolate(method="index", axis=1)
                    )
                    lead_var_val_in_key_timepoint = get_values_in_key_timepoint(
                        lead_var_interp
                    )

            lead_var_val_in_key_timepoint = lead_var_val_in_key_timepoint.timeseries()
            if not lead_var_val_in_key_timepoint.shape[1] == 1:  # pragma: no cover
                raise AssertionError(
                    "How did filtering for a single timepoint result in more than "
                    "one column?"
                )

            lead_var_val_in_key_timepoint = lead_var_val_in_key_timepoint.iloc[:, 0]

            scaling = data_follower_key_year_val / lead_var_val_in_key_timepoint
            output_ts = (lead_var.timeseries().T * scaling).T.reset_index()

            output_ts["variable"] = variable_follower
            output_ts["unit"] = data_follower_unit
            for col in in_iamdf.extra_cols:
                output_ts[col] = ""

            return IamDataFrame(output_ts)

        return filler

    def _get_iamdf_follower(self, variable_follower, variable_leaders):
        if len(variable_leaders) > 1:
            raise ValueError(
                "For `DatabaseCruncherLeadGas`, ``variable_leaders`` should only "
                "contain one variable"
            )

        self._check_follower_and_leader_in_db(variable_follower, variable_leaders)

        iamdf_follower = self._db.filter(variable=variable_follower)
        data_follower = iamdf_follower.data
        return iamdf_follower
